@@ -1,23 +1,27 @@
 #include "AudioIO.h"
+#include "BasicUI.h"
 #include "../Clipboard.h"
 #include "../CommonCommandFlags.h"
-#include "../LabelTrack.h"
+#include "LabelTrack.h"
 #include "Prefs.h"
 #include "Project.h"
 #include "ProjectAudioIO.h"
 #include "ProjectHistory.h"
-#include "../ProjectWindow.h"
+
 #include "../SelectUtilities.h"
 #include "SyncLock.h"
-#include "../TrackPanelAx.h"
+#include "TrackFocus.h"
 #include "../TrackPanel.h"
 #include "ViewInfo.h"
+#include "Viewport.h"
 #include "WaveTrack.h"
-#include "../commands/CommandContext.h"
-#include "../commands/CommandManager.h"
+#include "TimeStretching.h"
+#include "CommandContext.h"
 #include "../tracks/labeltrack/ui/LabelTrackView.h"
 #include "toolbars/ToolManager.h"
 
+#include <cassert>
+
 using Region = WaveTrack::Region;
 using Regions = WaveTrack::Regions;
 
@@ -42,7 +46,7 @@ const ReservedCommandFlag
             }
          );
       };
-      auto range = TrackList::Get( project ).Selected<const LabelTrack>()
+      auto range = TrackList::Get(project).Selected<const LabelTrack>()
          + test;
       return !range.empty();
    }
@@ -71,10 +75,11 @@ int DoAddLabel(
    const auto pFocusedTrack = trackFocus.Get();
 
    // Look for a label track at or after the focused track
+   auto begin = tracks.begin();
    auto iter = pFocusedTrack
       ? tracks.Find(pFocusedTrack)
-      : tracks.Any().begin();
-   auto lt = * iter.Filter< LabelTrack >();
+      : begin;
+   auto lt = * iter.Filter<LabelTrack>();
 
    // If none found, start a NEW label track and use it
    if (!lt)
@@ -98,7 +103,7 @@ int DoAddLabel(
          // Must remember the track to re-focus after finishing a label edit.
          // do NOT identify it by a pointer, which might dangle!  Identify
          // by position.
-         focusTrackNumber = pFocusedTrack->GetIndex();
+         focusTrackNumber = std::distance(begin, iter);
       }
       index =
          LabelTrackView::Get( *lt ).AddLabel(region, title, focusTrackNumber);
@@ -109,7 +114,7 @@ int DoAddLabel(
 
    if (!useDialog) {
       TrackFocus::Get(project).Set(lt);
-      lt->EnsureVisible();
+      Viewport::Get(project).ShowTrack(*lt);
    }
    trackPanel.SetFocus();
 
@@ -155,46 +160,84 @@ void GetRegionsByLabel(
    }
 }
 
-using EditFunction = std::function<void(Track *, double, double)>;
 
-//Executes the edit function on all selected wave tracks with
-//regions specified by selected labels
-//If No tracks selected, function is applied on all tracks
-//If the function replaces the selection with audio of a different length,
+auto GetTracksToEditByLabel(const AudacityProject& project, const TrackList& tracks)
+{
+   const bool syncLocked = SyncLockState::Get(project).IsSyncLocked();
+   const bool playableTrackSelected = !tracks.Selected<const PlayableTrack>().empty();
+
+   return [syncLocked, playableTrackSelected](const auto pTrack) {
+      // All sync-locked tracks should be edited,
+      // if sync lock is enabled regardless of selection
+      if (syncLocked) {
+         return SyncLock::IsSyncLockSelected(*pTrack);
+      }
+
+      // Non-playable tracks should not be edited if sync-lock disabled
+      if (dynamic_cast<const PlayableTrack*>(pTrack) == nullptr) {
+         return false;
+      }
+
+      // If no playable tracks were selected,
+      // then all tracks should be edited
+      return playableTrackSelected ? pTrack->IsSelected() : true;
+   };
+}
+
+using EditFunction = std::function<void(Track& track, double, double)>;
+using EditFunctionWithProgress =
+   std::function<void(Track& track, double, double, ProgressReporter)>;
+
+// Executes the edit function on all selected wave tracks with
+// regions specified by selected labels
+// If No tracks selected, function is applied on all tracks
+// If the function replaces the selection with audio of a different length,
 // bSyncLockedTracks should be set true to perform the same action on sync-lock
 // selected tracks.
-void EditByLabel(AudacityProject &project,
-   TrackList &tracks, const SelectedRegion &selectedRegion,
-   EditFunction action)
+// If `progress` is non-null, `action` is passed a callable `ProgressReporter`.
+void EditByLabel(
+   AudacityProject& project, TrackList& tracks,
+   const SelectedRegion& selectedRegion, EditFunctionWithProgress action,
+   ProgressReporter progress)
 {
    Regions regions;
 
-   GetRegionsByLabel( tracks, selectedRegion, regions );
-   if( regions.size() == 0 )
+   GetRegionsByLabel(tracks, selectedRegion, regions);
+   if (regions.empty())
       return;
 
-   const bool notLocked = (!SyncLockState::Get(project).IsSyncLocked() &&
-                           (tracks.Selected<PlayableTrack>()).empty());
+   const auto tracksToEdit = tracks.Any<Track>() + GetTracksToEditByLabel(project, tracks);
 
-   //Apply action on tracks starting from
-   //labeled regions in the end. This is to correctly perform
-   //actions like 'Delete' which collapse the track area.
-   for (auto t : tracks.Any())
-   {
-      const bool playable = dynamic_cast<const PlayableTrack *>(t) != nullptr;
-
-      if (SyncLock::IsSyncLockSelected(t) || notLocked && playable)
-      {
-         for (int i = (int)regions.size() - 1; i >= 0; i--)
-         {
-            const Region &region = regions.at(i);
-            action(t, region.start, region.end);
-         }
-      }
-   }
+   BasicUI::SplitProgress(
+      tracksToEdit.begin(), tracksToEdit.end(),
+      [&](Track* aTrack, const ProgressReporter& child) {
+         // Apply action on tracks starting from labeled regions in the end.
+         // This is to correctly perform actions like 'Delete' which collapse
+         // the track area.
+         BasicUI::SplitProgress(
+            regions.rbegin(), regions.rend(),
+            [&](const Region& region, const ProgressReporter& grandChild) {
+               action(*aTrack, region.start, region.end, grandChild);
+            },
+            child);
+      },
+      progress);
 }
 
-using EditDestFunction = std::function<Track::Holder (Track *, double, double)>;
+void EditByLabel(
+   AudacityProject& project, TrackList& tracks,
+   const SelectedRegion& selectedRegion, EditFunction action)
+{
+   EditByLabel(
+      project, tracks, selectedRegion,
+      [&](Track& track, double t0, double t1, ProgressReporter) {
+         action(track, t0, t1);
+      },
+      nullptr);
+}
+
+using EditDestFunction =
+   std::function<Track::Holder(Track &, double, double)>;
 
 //Executes the edit function on all selected wave tracks with
 //regions specified by selected labels
@@ -202,73 +245,66 @@ using EditDestFunction = std::function<Track::Holder (Track *, double, double)>;
 //Functions copy the edited regions to clipboard, possibly in multiple tracks
 //This probably should not be called if *action() changes the timeline, because
 // the copy needs to happen by track, and the timeline change by group.
-void EditClipboardByLabel( AudacityProject &project,
+void EditClipboardByLabel(AudacityProject &project,
    TrackList &tracks, const SelectedRegion &selectedRegion,
-   EditDestFunction action )
+   EditDestFunction action)
 {
    Regions regions;
 
-   GetRegionsByLabel( tracks, selectedRegion, regions );
-   if( regions.size() == 0 )
+   GetRegionsByLabel(tracks, selectedRegion, regions);
+   if (regions.empty())
       return;
 
-   const bool notLocked = (!SyncLockState::Get(project).IsSyncLocked() &&
-                           (tracks.Selected<PlayableTrack>()).empty());
+   const auto tracksToEdit = tracks.Any<Track>() + GetTracksToEditByLabel(project, tracks);
 
    auto &clipboard = Clipboard::Get();
    clipboard.Clear();
 
-   auto pNewClipboard = TrackList::Create( nullptr );
+   auto pNewClipboard = TrackList::Create(nullptr);
    auto &newClipboard = *pNewClipboard;
 
    //Apply action on wavetracks starting from
    //labeled regions in the end. This is to correctly perform
    //actions like 'Cut' which collapse the track area.
 
-   for( auto t : tracks.Any())
-   {
-      const bool playable = dynamic_cast<const PlayableTrack *>(t) != nullptr;
+   for (auto t : tracksToEdit) {
+      // These tracks accumulate the needed clips, right to left:
+      Track::Holder merged;
+      for (size_t i = regions.size(); i--;) {
+         const Region &region = regions.at(i);
+         if (auto dest = action(*t, region.start, region.end)) {
+            if (!merged)
+               merged = dest;
+            else {
+               // Paste to the beginning; unless this is the first region,
+               // offset the track to account for time between the regions
+               if (i + 1 < regions.size())
+                  merged->ShiftBy(
+                     regions.at(i + 1).start - region.end);
 
-      if (SyncLock::IsSyncLockSelected(t) || notLocked && playable)
-      {
-         // This track accumulates the needed clips, right to left:
-         Track::Holder merged;
-         for( int i = (int)regions.size() - 1; i >= 0; i-- )
-         {
-            const Region &region = regions.at(i);
-            auto dest = action(t, region.start, region.end );
-            if( dest )
-            {
-               Track::FinishCopy( t, dest.get() );
-               if( !merged )
-                  merged = dest;
-               else
-               {
-                  // Paste to the beginning; unless this is the first region,
-                  // offset the track to account for time between the regions
-                  if (i < (int)regions.size() - 1)
-                     merged->Offset(
-                        regions.at(i + 1).start - region.end);
+               // If GetEditClipsCanMove is not set, we need to manually
+               // shift the merged track by the end time of the dest clip,
+               // to make room for pasting because Paste does not
+               // handle shifting in this case.
+               if (!GetEditClipsCanMove())
+                  merged->ShiftBy(dest->GetEndTime());
 
-                  // dest may have a placeholder clip at the end that is
-                  // removed when pasting, which is okay because we proceed
-                  // right to left.  Any placeholder already in merged is kept.
-                  // Only the rightmost placeholder is important in the final
-                  // result.
-                  merged->Paste( 0.0 , dest.get() );
-               }
+               // dest may have a placeholder clip at the end that is
+               // removed when pasting, which is okay because we proceed
+               // right to left.  Any placeholder already in merged is kept.
+               // Only the rightmost placeholder is important in the final
+               // result.
+               merged->Paste(0.0, *dest);
             }
-            else
-               // nothing copied but there is a 'region', so the 'region' must
-               // be a 'point label' so offset
-               if (i < (int)regions.size() - 1)
-                  if (merged)
-                     merged->Offset(
-                        regions.at(i + 1).start - region.end);
          }
-         if( merged )
-            newClipboard.Add( merged );
+         else
+            // nothing copied but there is a 'region', so the 'region' must
+            // be a 'point label' so offset
+            if (i + 1 < regions.size() && merged)
+               merged->ShiftBy(regions.at(i + 1).start - region.end);
       }
+      if (merged)
+         newClipboard.Add(merged);
    }
 
    // Survived possibility of exceptions.  Commit changes to the clipboard now.
@@ -322,13 +358,13 @@ void OnPasteNewLabel(const CommandContext &context)
    bool bPastedSomething = false;
 
    {
-      auto trackRange = tracks.Selected< const LabelTrack >();
+      auto trackRange = tracks.Selected<const LabelTrack>();
       if (trackRange.empty())
       {
          // If there are no selected label tracks, try to choose the first label
          // track after some other selected track
          Track *t = *tracks.Selected().begin()
-            .Filter( &Track::Any )
+            .Filter(&Track::Any)
             .Filter<LabelTrack>();
 
          // If no match found, add one
@@ -341,7 +377,7 @@ void OnPasteNewLabel(const CommandContext &context)
    }
 
    LabelTrack *plt = NULL; // the previous track
-   for ( auto lt : tracks.Selected< LabelTrack >() )
+   for ( auto lt : tracks.Selected<LabelTrack>() )
    {
       // Unselect the last label, so we'll have just one active label when
       // we're done
@@ -364,7 +400,7 @@ void OnPasteNewLabel(const CommandContext &context)
    // and set focus
    if (plt) {
       TrackFocus::Get(project).Set(plt);
-      plt->EnsureVisible();
+      Viewport::Get(project).ShowTrack(*plt);
       trackPanel.SetFocus();
    }
 
@@ -386,45 +422,32 @@ void OnToggleTypeToCreateLabel(const CommandContext &WXUNUSED(context) )
 void OnCutLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
      return;
 
    // Because of grouping the copy may need to operate on different tracks than
    // the clear, so we do these actions separately.
-   auto copyfunc = [&](Track *track, double t0, double t1)
-   {
-      Track::Holder dest = nullptr;
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
-            dest = t->CopyNonconst(t0, t1);
-         }
-      );
-      return dest;
+   auto copyfunc = [&](Track &track, double t0, double t1) {
+      Track::Holder result;
+      track.TypeSwitch( [&](WaveTrack &wt) { result = wt.Copy(t0, t1); } );
+      return result;
    };
-   EditClipboardByLabel( project, tracks, selectedRegion, copyfunc );
+   EditClipboardByLabel(project, tracks, selectedRegion, copyfunc);
 
    bool enableCutlines = gPrefs->ReadBool(wxT( "/GUI/EnableCutLines"), false);
-   auto editfunc = [&](Track *track, double t0, double t1)
-   {
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
+   auto editfunc = [&](Track &track, double t0, double t1) {
+      track.TypeSwitch(
+         [&](WaveTrack &t) {
             if (enableCutlines)
-            {
-               t->ClearAndAddCutLine(t0, t1);
-            }
+               t.ClearAndAddCutLine(t0, t1);
             else
-            {
-               t->Clear(t0, t1);
-            }
+               t.Clear(t0, t1);
          },
-         [&](Track *t)
-         {
-            t->Clear(t0, t1);
+         [&](Track &t) {
+            t.Clear(t0, t1);
          }
       );
    };
@@ -432,243 +455,218 @@ void OnCutLabels(const CommandContext &context)
 
    selectedRegion.collapseToT0();
 
-   ProjectHistory::Get( project ).PushState(
+   ProjectHistory::Get(project).PushState(
    /* i18n-hint: (verb) past tense.  Audacity has just cut the labeled audio
       regions.*/
-      XO( "Cut labeled audio regions to clipboard" ),
+      XO("Cut labeled audio regions to clipboard"),
    /* i18n-hint: (verb)*/
-      XO( "Cut Labeled Audio" ) );
+      XO("Cut Labeled Audio"));
 }
 
 void OnDeleteLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
       return;
 
-   auto editfunc = [&](Track *track, double t0, double t1)
-   {
-      track->Clear(t0, t1);
+   auto editfunc = [&](Track &track, double t0, double t1) {
+      track.TypeSwitch( [&](Track &t) { t.Clear(t0, t1); } );
    };
    EditByLabel(project, tracks, selectedRegion, editfunc);
 
    selectedRegion.collapseToT0();
 
-   ProjectHistory::Get( project ).PushState(
+   ProjectHistory::Get(project).PushState(
       /* i18n-hint: (verb) Audacity has just deleted the labeled audio regions*/
-      XO( "Deleted labeled audio regions" ),
+      XO("Deleted labeled audio regions"),
       /* i18n-hint: (verb)*/
-      XO( "Delete Labeled Audio" ) );
+      XO("Delete Labeled Audio"));
 }
 
 void OnSplitCutLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
       return;
 
-   auto copyfunc = [&](Track *track, double t0, double t1)
-   {
-      Track::Holder dest = nullptr;
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
-            dest = t->SplitCut(t0, t1);
+   auto copyfunc = [&](Track &track, double t0, double t1) {
+      Track::Holder result;
+      track.TypeSwitch(
+         [&](WaveTrack &wt) {
+            result = wt.SplitCut(t0, t1);
          },
-         [&](Track *t)
-         {
-            dest = t->Copy(t0, t1);
-            t->Silence(t0, t1);
+         [&](Track &t) {
+            result = t.Copy(t0, t1);
+            t.Silence(t0, t1);
          }
       );
-      return dest;
+      return result;
    };
-   EditClipboardByLabel( project, tracks, selectedRegion, copyfunc );
+   EditClipboardByLabel(project, tracks, selectedRegion, copyfunc);
 
-   ProjectHistory::Get( project ).PushState(
-      /* i18n-hint: (verb) Audacity has just split cut the labeled audio
-         regions*/
-      XO( "Split Cut labeled audio regions to clipboard" ),
-      /* i18n-hint: (verb) Do a special kind of cut on the labels*/
-      XO( "Split Cut Labeled Audio" ) );
+   ProjectHistory::Get(project).PushState(
+      XO("Cut labeled audio regions to clipboard and leave gap"),
+      XO("Cut labeled audio and leave gap"));
 }
 
 void OnSplitDeleteLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
       return;
 
-   auto editfunc = [&](Track *track, double t0, double t1)
-   {
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
-            t->SplitDelete(t0, t1);
+   auto editfunc = [&](Track &track, double t0, double t1) {
+      track.TypeSwitch(
+         [&](WaveTrack &t) {
+            t.SplitDelete(t0, t1);
          },
-         [&](Track *t)
-         {
-            t->Silence(t0, t1);
+         [&](Track &t) {
+            t.Silence(t0, t1);
          }
       );
    };
    EditByLabel(project, tracks, selectedRegion, editfunc);
 
-   ProjectHistory::Get( project ).PushState(
+   ProjectHistory::Get(project).PushState(
       /* i18n-hint: (verb) Audacity has just done a special kind of DELETE on
          the labeled audio regions */
-      XO( "Split Deleted labeled audio regions" ),
+      XO("Split Deleted labeled audio regions"),
       /* i18n-hint: (verb) Do a special kind of DELETE on labeled audio
          regions */
-      XO( "Split Delete Labeled Audio" ) );
+      XO("Split Delete Labeled Audio"));
 }
 
 void OnSilenceLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
       return;
 
-   auto editfunc = [&](Track *track, double t0, double t1)
-   {
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
-            t->Silence(t0, t1);
-         }
-      );
+   auto editfunc = [&](Track &track, double t0, double t1) {
+      // TODO use progress-bar utilities pending in
+      // https://github.com/audacity/audacity/pull/5043
+      track.TypeSwitch([&](WaveTrack& t) { t.Silence(t0, t1, {}); });
    };
    EditByLabel(project, tracks, selectedRegion, editfunc);
 
-   ProjectHistory::Get( project ).PushState(
+   ProjectHistory::Get(project).PushState(
       /* i18n-hint: (verb)*/
-      XO( "Silenced labeled audio regions" ),
+      XO("Silenced labeled audio regions"),
       /* i18n-hint: (verb)*/
-      XO( "Silence Labeled Audio" ) );
+      XO("Silence Labeled Audio"));
 }
 
 void OnCopyLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
       return;
 
-   auto copyfunc = [&](Track *track, double t0, double t1)
-   {
-      Track::Holder dest = nullptr;
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
-            dest = t->CopyNonconst(t0, t1);
-         });
-      return dest;
+   auto copyfunc = [&](Track &track, double t0, double t1) {
+      Track::Holder result;
+      track.TypeSwitch( [&](WaveTrack &wt) { result = wt.Copy(t0, t1); } );
+      return result;
    };
-   EditClipboardByLabel( project, tracks, selectedRegion, copyfunc );
+   EditClipboardByLabel(project, tracks, selectedRegion, copyfunc);
 
-   ProjectHistory::Get( project ).PushState( XO( "Copied labeled audio regions to clipboard" ),
-   /* i18n-hint: (verb)*/
-      XO( "Copy Labeled Audio" ) );
+   ProjectHistory::Get(project)
+      .PushState(XO("Copied labeled audio regions to clipboard"),
+         /* i18n-hint: (verb)*/
+         XO("Copy Labeled Audio"));
 }
 
 void OnSplitLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
       return;
 
-   auto editfunc = [&](Track *track, double t0, double t1)
-   {
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
-            t->Split(t0, t1);
-         }
-      );
+   auto editfunc = [&](Track &track, double t0, double t1) {
+      track.TypeSwitch( [&](WaveTrack &t) { t.Split(t0, t1); } );
    };
    EditByLabel(project, tracks, selectedRegion, editfunc);
 
-   ProjectHistory::Get( project ).PushState(
+   ProjectHistory::Get(project).PushState(
       /* i18n-hint: (verb) past tense.  Audacity has just split the labeled
          audio (a point or a region)*/
-      XO( "Split labeled audio (points or regions)" ),
+      XO("Split labeled audio (points or regions)"),
       /* i18n-hint: (verb)*/
-      XO( "Split Labeled Audio" ) );
+      XO("Split Labeled Audio"));
 }
 
 void OnJoinLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
       return;
 
-   auto editfunc = [&](Track *track, double t0, double t1)
-   {
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
-            t->Join(t0, t1);
-         }
-      );
+   auto editfunc = [&](Track& track, double t0, double t1, ProgressReporter reportProgress) {
+      track.TypeSwitch(
+         [&](WaveTrack& t) {
+            // Extend Join Region by one sample, to ensure
+            // that Split-Join operation on a given label is reversible and consistent
+            const auto sampleTime = 1.0 / t.GetRate();
+            t.Join(t0 - sampleTime, t1 + sampleTime, std::move(reportProgress));
+         });
    };
-   EditByLabel(project, tracks, selectedRegion, editfunc);
+   TimeStretching::WithClipRenderingProgress(
+      [&](ProgressReporter progress) {
+         EditByLabel(project, tracks, selectedRegion, editfunc, progress);
+      });
 
-   ProjectHistory::Get( project ).PushState(
+   ProjectHistory::Get(project).PushState(
       /* i18n-hint: (verb) Audacity has just joined the labeled audio (points or
          regions) */
-      XO( "Joined labeled audio (points or regions)" ),
+      XO("Joined labeled audio (points or regions)"),
       /* i18n-hint: (verb) */
-      XO( "Join Labeled Audio" ) );
+      XO("Join Labeled Audio"));
 }
 
 void OnDisjoinLabels(const CommandContext &context)
 {
    auto &project = context.project;
-   auto &tracks = TrackList::Get( project );
-   auto &selectedRegion = ViewInfo::Get( project ).selectedRegion;
+   auto &tracks = TrackList::Get(project);
+   auto &selectedRegion = ViewInfo::Get(project).selectedRegion;
 
-   if( selectedRegion.isPoint() )
+   if (selectedRegion.isPoint())
       return;
 
-   auto editfunc = [&](Track *track, double t0, double t1)
-   {
-      track->TypeSwitch(
-         [&](WaveTrack *t)
-         {
-            wxBusyCursor busy;
-            t->Disjoin(t0, t1);
-         }
-      );
+   auto editfunc = [&](Track &track, double t0, double t1) {
+      track.TypeSwitch( [&](WaveTrack &t) {
+         wxBusyCursor busy;
+         t.Disjoin(t0, t1);
+      } );
    };
    EditByLabel(project, tracks, selectedRegion, editfunc);
 
-   ProjectHistory::Get( project ).PushState(
+   ProjectHistory::Get(project).PushState(
       /* i18n-hint: (verb) Audacity has just detached the labeled audio regions.
       This message appears in history and tells you about something
       Audacity has done.*/
-      XO( "Detached labeled audio regions" ),
+      XO("Detached labeled audio regions"),
       /* i18n-hint: (verb)*/
-      XO( "Detach Labeled Audio" ) );
+      XO("Detach Labeled Audio"));
 }
 
 void OnNewLabelTrack(const CommandContext &context)
@@ -686,29 +684,28 @@ void OnNewLabelTrack(const CommandContext &context)
       .PushState(XO("Created new label track"), XO("New Track"));
 
    TrackFocus::Get(project).Set(track);
-   track->EnsureVisible();
+   Viewport::Get(project).ShowTrack(*track);
 }
 
 // Menu definitions
 
-using namespace MenuTable;
-BaseItemSharedPtr LabelEditMenus()
+using namespace MenuRegistry;
+auto LabelEditMenus()
 {
-   using namespace MenuTable;
-   using Options = CommandManager::Options;
+   using namespace MenuRegistry;
 
    static const auto NotBusyLabelsAndWaveFlags =
       AudioIONotBusyFlag() |
       LabelsSelectedFlag() | WaveTracksExistFlag() | TimeSelectedFlag();
 
    // Returns TWO menus.
-   
-   static BaseItemSharedPtr menus{
+
+   static auto menus = std::shared_ptr{
    Items( wxT("LabelEditMenus"),
-   
+
    Menu( wxT("Labels"), XXO("&Labels"),
       Section( "",
-         Command( wxT("EditLabels"), XXO("&Edit Labels..."), OnEditLabels,
+         Command( wxT("EditLabels"), XXO("Label &Editor"), OnEditLabels,
                     AudioIONotBusyFlag() )
       ),
 
@@ -731,7 +728,7 @@ BaseItemSharedPtr LabelEditMenus()
 
       Section( "",
          Command( wxT("TypeToCreateLabel"),
-            XXO("&Type to Create a Label (on/off)"),
+            XXO("&Typing Creates New Labels"),
             OnToggleTypeToCreateLabel, AlwaysEnabledFlag,
             Options{}.CheckTest(wxT("/GUI/TypeToCreateLabel"), false) )
       )
@@ -745,21 +742,21 @@ BaseItemSharedPtr LabelEditMenus()
          Command( wxT("CutLabels"), XXO("&Cut"), OnCutLabels,
             AudioIONotBusyFlag() | LabelsSelectedFlag() | WaveTracksExistFlag() |
                TimeSelectedFlag(),
-               Options{ wxT("Alt+X"), XO("Label Cut") } ),
+               Options{ wxT("Alt+X"), XO("Cut by label") } ),
          Command( wxT("DeleteLabels"), XXO("&Delete"), OnDeleteLabels,
             AudioIONotBusyFlag() | LabelsSelectedFlag() | WaveTracksExistFlag() |
                TimeSelectedFlag(),
-            Options{ wxT("Alt+K"), XO("Label Delete") } )
+            Options{ wxT("Alt+K"), XO("Delete by label") } )
       ),
 
       Section( "",
          /* i18n-hint: (verb) A special way to cut out a piece of audio*/
-         Command( wxT("SplitCutLabels"), XXO("&Split Cut"),
+         Command( wxT("SplitCutLabels"), XXO("C&ut and leave gap"),
             OnSplitCutLabels, NotBusyLabelsAndWaveFlags,
-            Options{ wxT("Alt+Shift+X"), XO("Label Split Cut") } ),
-         Command( wxT("SplitDeleteLabels"), XXO("Sp&lit Delete"),
+            Options{ wxT("Alt+Shift+X"), XO("Cut by labels and leave gap") } ),
+         Command( wxT("SplitDeleteLabels"), XXO("De&lete and leave gap"),
             OnSplitDeleteLabels, NotBusyLabelsAndWaveFlags,
-            Options{ wxT("Alt+Shift+K"), XO("Label Split Delete") } )
+            Options{ wxT("Alt+Shift+K"), XO("Delete by labels and leave gap") } )
       ),
 
       Section( "",
@@ -791,15 +788,15 @@ BaseItemSharedPtr LabelEditMenus()
    return menus;
 }
 
-AttachedItem sAttachment1{
+AttachedItem sAttachment1{ Indirect(LabelEditMenus()),
    { wxT("Edit/Other"),
-     { OrderingHint::Before, wxT("EditMetaData") } },
-   Shared( LabelEditMenus() )
+     { OrderingHint::Before, wxT("EditMetaData") } }
 };
 
-AttachedItem sAttachment2{ wxT("Tracks/Add/Add"),
+AttachedItem sAttachment2{
    Command( wxT("NewLabelTrack"), XXO("&Label Track"),
-      OnNewLabelTrack, AudioIONotBusyFlag() )
+      OnNewLabelTrack, AudioIONotBusyFlag() ),
+   wxT("Tracks/Add/Add")
 };
 
 }
